from typing import Tuple, List, Dict, Union

from rest_framework.reverse import reverse
from bridger.metadata.mixins import BridgerViewSetConfig


class EndpointConfig(BridgerViewSetConfig):
    """ Configuration for endpoints to determine which buttons and actions are available

    In Instance Views:
    - instance_endpoint -> Save Button [PATCH|PUT]
    - list_endpoint -> Return to list?
    - delete_endpoint -> Delete Button (Instance) [DELETE]
    - create_endpoint -> Create Button [POST]

    In List Views:
    - instance_endpoint -> Makes instance clickable and opens existing instance [GET] !!With Handlebar for PK
    - list_endpoint -> Means nothing?
    - delete_endpoint -> Delete Button (Batch Delete) [DELETE] !!With Handlebar for PK (not really?)
    - create_endpoint -> Create Buttons [POST]

    """

    PK_FIELD = "id"

    def get_endpoint(self, **kwargs):
        model = self.view.get_model()
        if basename_method := getattr(model, "get_endpoint_basename", None):
            basename = basename_method()
            is_list = kwargs.get("is_list", False)
            if self.instance and not is_list:
                return reverse(f"{basename}-detail", args=[self.view.kwargs.get("pk")], request=self.request)
            else:
                return reverse(f"{basename}-list", request=self.request)
        return None

    def get_instance_endpoint(self, **kwargs):
        return self.get_endpoint()

    def _get_instance_endpoint(self):
        if endpoint := self.get_instance_endpoint():
            from bridger.viewsets import ViewSet
            content_type = self.view.get_content_type()
            if content_type:
                change_permission = f"{content_type.app_label}.change_{content_type.model}"
                permission = self.request.user.has_perm(change_permission)
            elif isinstance(self.view, ViewSet):
                permission = True
            read_only = getattr(self.view, "READ_ONLY", False)
            if self.instance and permission and not read_only:
                return endpoint 
            elif not self.instance:
                pk_identifier = "{{"+ self.PK_FIELD +"}}/"
                return endpoint + pk_identifier

        return None

    def get_list_endpoint(self, **kwargs):
        return self.get_endpoint(is_list=True)

    def _get_list_endpoint(self):
        return self.get_list_endpoint()

    def get_delete_endpoint(self, **kwargs):
        return self.get_endpoint()

    def _get_delete_endpoint(self):
        read_only = getattr(self.view, "READ_ONLY", False)
        if read_only:
            return None

        if endpoint := self.get_delete_endpoint():
            content_type = self.view.get_content_type()
            delete_permission = f"{content_type.app_label}.delete_{content_type.model}"
            if self.instance and self.request.user.has_perm(delete_permission):
                return endpoint 
            elif not self.instance:
                pk_identifier = "{{"+ self.PK_FIELD +"}}/"
                return endpoint + pk_identifier
            else:
                return endpoint
        return None

    def get_create_endpoint(self, **kwargs):
        return self.get_endpoint(is_list=True)

    def _get_create_endpoint(self):
        from bridger.viewsets import ViewSet
        read_only = getattr(self.view, "READ_ONLY", False)
        if read_only:
            return None

        if endpoint := self.get_create_endpoint():
            content_type = self.view.get_content_type()
            if content_type:
                create_permission = f"{content_type.app_label}.add_{content_type.model}"
                permission = self.request.user.has_perm(create_permission)
            elif isinstance(self.view, ViewSet):
                permission = True
            if permission:
                return endpoint
            
        return None
  
    def get_metadata(self) -> Dict:
        yield "instance", self._get_instance_endpoint()
        yield "list", self._get_list_endpoint()
        yield "delete", self._get_delete_endpoint()
        yield "create", self._get_create_endpoint()

    @classmethod
    def get_metadata_key(cls):
        return "endpoints"

    @classmethod
    def get_metadata_fieldname(cls):
        return "endpoint_config_class"

class EndpointConfigMixin:
    endpoint_config_class = EndpointConfig


import logging
from datetime import datetime

from rest_framework.fields import empty
from rest_framework.settings import api_settings

logger = logging.getLogger(__name__)


class BridgerSerializerFieldMixin:
    """
    A mixin that takes care of adding all the necessary magic to each implementation
    of the serializer fields
    """

    def __init__(self, *args, **kwargs):
        self.extra = kwargs.pop("extra", None)
        self.decorators = kwargs.pop("decorators", [])
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name):

        if meta := getattr(self.parent, "Meta", None):
            if field_name in getattr(meta, "required_fields", []):
                self.required = True

        representation = {
            "key": field_name,
            "label": getattr(self, "label", None),
            "type": getattr(self, "field_type", "undefined"),
            "required": getattr(self, "required", True),
            "read_only": getattr(self, "read_only", False),
        }

        default = getattr(self, "default", None)

        if default and default != empty:
            if isinstance(default, datetime):
                default = default.strftime(api_settings.DATETIME_FORMAT)

            representation["default"] = default() if callable(default) else default
        else:
            try:
                default = self.parent.Meta.model._meta._forward_fields_map[field_name].default
                if isinstance(default, (str, float, int)):
                    representation["default"] = default
            except:  # TODO Add some explicit exception handling
                pass

        representation["decorators"] = getattr(self, "decorators", [])

        for _attr in ["help_text", "extra"]:
            attr = getattr(self, _attr, None)
            if attr:
                representation[_attr] = attr

        return representation



from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from bridger.settings import bridger_settings

from .registry import default_registry


class MenuAPIView(APIView):
    @property
    def permission_classes(self):
        bridger_auth = bridger_settings.DEFAULT_AUTH_CONFIG(None)

        if bridger_auth["type"] is None:
            return []

        return [IsAuthenticated]

    def get(self, request: Request) -> Response:
        default_registry.request = request
        return Response(list(default_registry))



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig
from bridger.filters import DjangoFilterBackend
from bridger.utils import ilen

class FilterFieldsBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if DjangoFilterBackend in getattr(self.view, "filter_backends", []):
            filterset_class = DjangoFilterBackend().get_filterset_class(self.view, self.view.queryset)
            if hasattr(filterset_class, "base_filters"):
                filterset = filterset_class.base_filters.items()
                field_names = [v.field_name for k, v in filterset_class.base_filters.items()]
                stored_fields = dict()
                for index, (name, field) in enumerate(filterset):

                    upcomming_fields = ilen(filter(lambda f: f == field.field_name, field_names[index + 1 :]))
                    representation = field.get_representation(self.request, name, self.view)

                    if upcomming_fields > 0 or field.field_name in stored_fields:
                        _representation = stored_fields.get(field.field_name, representation)
                        _representation["lookup_expr"].update(representation["lookup_expr"])
                        _representation["default"].update(representation["default"])
                        stored_fields[field.field_name] = _representation

                    if upcomming_fields == 0:
                        yield field.field_name, stored_fields.get(field.field_name, representation)

    @classmethod
    def get_metadata_key(cls):
        return "filter_fields"



from .configs import (
    DocumentationBridgerViewSetConfig,
    FieldsBridgerViewSetConfig,
    IdentifierBridgerViewSetConfig,
    PaginationBridgerViewSetConfig,
    PKBridgerViewSetConfig,
    TypeBridgerViewSetConfig,
    FilterFieldsBridgerViewSetConfig,
    SearchFieldsBridgerViewSetConfig,
    OrderingFieldsBridgerViewSetConfig,
)
from .metadata import BridgerMetadata



# Generated by Django 3.0.6 on 2020-05-14 15:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0010_asset_content_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="asset", name="file_url_name", field=models.CharField(blank=True, max_length=1024, null=True),
        ),
    ]



from calendar import monthrange
from datetime import date

from django.utils.timezone import localdate

from bridger.utils.date import get_end_date_from_date, get_start_date_from_date


def current_quarter_date_start(field, request, view):
    return get_start_date_from_date(localdate())


def current_quarter_date_end(field, request, view):
    return get_end_date_from_date(localdate())


def current_month_date_start(field, request, view):
    today = date.today()
    return date(today.year, today.month, 1)


def current_month_date_end(field, request, view):
    today = date.today()
    return date(today.year, today.month, monthrange(today.year, today.month)[1])


def current_quarter_date_interval(field, request, view):
    return (
        current_quarter_date_start(field, request, view),
        current_quarter_date_end(field, request, view),
    )



from django.apps import apps
import importlib 
import factory
from rest_framework.test import APIRequestFactory
from rest_framework.request import Request
from rest_framework.authtoken.models import Token
from django.urls import reverse
import json

from functools import partial
from typing import Any, Dict

from factory import Factory
from factory.base import StubObject

from django.db import models
from bridger.tests.signals import add_kwargs, add_data_factory

import datetime

def get_all_subclasses(klass):
    for subclass in klass.__subclasses__():
        yield subclass
        yield from get_all_subclasses(subclass)

def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])

def get_model_factory(model):
    # mf = [cls for cls in factory.django.DjangoModelFactory.__subclasses__() if cls._meta.model == model]
    mfs = [cls for cls in all_subclasses(factory.django.DjangoModelFactory) if cls._meta.model == model]
    if is_empty(mfs):
        return None
    else:
        mf = [mf for mf in mfs if mf.__name__== model.__name__+"Factory"]  
        if not is_empty(mf):
            return mf[0]
        else:      
            return mfs[0]

def get_data_factory_mvs(obj, mvs, delete=False, update=False, superuser=None):
    request = APIRequestFactory().get("")
    if superuser:
        request.user = superuser
        request.parser_context = {"view": mvs}
        serializer = mvs().serializer_class(obj, context= {'request': request})
    else:
        serializer = mvs().serializer_class(obj, context= {'request': Request(request)})

    # fields_models = [m.name for m in mvs().serializer_class.Meta.model._meta.get_fields()]
    dict_fields_models = {}
    for m in mvs().serializer_class.Meta.model._meta.get_fields():
        dict_fields_models[m.name] = m

    data = {}
    for key, value in serializer.data.items():
        if key in dict_fields_models.keys() and key != "frontend_settings":
            if key == "auth_token":
                data[key], created = Token.objects.get_or_create(user=obj)
            elif dict_fields_models[key].get_internal_type() in ["FileField", "ImageField", "JSONField"] or key == "content_type" :
                # data[key] = open(value.replace("http://testserver/",""), 'rb')
                pass
            else:
                data[key] = value
        
    if (delete and superuser) or (update and superuser):
        kwargs = {"superuser": superuser, "obj": obj}
        remote_data_factory = add_data_factory.send(mvs, **kwargs)
        if remote_data_factory :
            _, r_datakwarg = remote_data_factory[0]
            data.update(r_datakwarg)
            superuser = r_datakwarg["superuser"]

    # delete object for create with post method
    if delete:
        mvs().serializer_class.Meta.model.objects.filter(pk = obj.pk).delete()

    return data, superuser

def get_kwargs(obj, mvs, request, data=None):
    fields_models = [m for m in mvs().serializer_class.Meta.model._meta.get_fields()] 
    mvs.kwargs = {}
    request.parser_context = {"view": mvs}
    
    serializer = mvs().serializer_class(obj, context= {'request': request, "view": mvs})
    
    # serializer = mvs().get_serializer(obj)
    kwargs = {}
    for field in fields_models:
        if field.is_relation and field.name in serializer.data.keys():
            
            if is_empty(serializer.data[field.name]) and data:          
                kwargs[field.name+"_id"] = data[field.name]
            else:
                kwargs[field.name+"_id"] = serializer.data[field.name]
    # inject the profile for test the view ex.(InChargeCustomerModelViewSet) of crm
    if hasattr(request.user, 'profile'):
        kwargs["profile"] = request.user.profile
        kwargs["user"] = request.user
    if data:
        kwargs["data"] = data
    kwargs["obj_factory"] = obj
    remote_kwargs = add_kwargs.send(mvs, **kwargs)
    if remote_kwargs:
        _, r_kwarg = remote_kwargs[0]
        kwargs.update(r_kwarg)  
    return kwargs


def generate_dict_factory(factory: Factory):
    def convert_dict_from_stub(stub: StubObject) -> Dict[str, Any]:
        stub_dict = stub.__dict__
        for key, value in stub_dict.items():
            if isinstance(value, StubObject):
                stub_dict[key] = convert_dict_from_stub(value)
        return stub_dict

    def dict_factory(factory, **kwargs):
        stub = factory.stub(**kwargs)
        stub_dict = convert_dict_from_stub(stub)
        return stub_dict

    return partial(dict_factory, factory)

# check if an element of a list is present in the other
contains = lambda lst1, lst2: any( elt in lst1  for elt in lst2 )

def is_intermediate_table_m2m(model):
    if "_" in model.__qualname__ and not is_empty(model._meta.unique_together) :
        list_models = [elt.lower() for elt in model.__qualname__.split("_")]
        list_models_id = [elt.lower()+"_id" for elt in model.__qualname__.split("_")]
        if (contains(list_models, model._meta.unique_together[0]) and
            contains(list_models, model.__dict__.keys()) and
            contains(list_models_id, model.__dict__.keys()) ):
            return True
    return False

        # print(model.__dict__.keys())
        # print(model._meta.__dict__)
        # print(list_models)
        # print(model._meta.unique_together)
      

def is_empty(any_structure):
    if any_structure:
        return False
    else:
        return True

def get_factory_custom_user():
    list_models = [m for m in get_all_subclasses(models.Model) if m.__name__ == ('User') and not m.__module__.startswith(('bridger', 'django', 'rest_framework', 'dynamic_preferences', 'eventtools')) 
                and not m._meta.abstract and not is_intermediate_table_m2m(m)] 

    userfactory = [get_model_factory(modl) for modl in list(dict.fromkeys(list_models)) if get_model_factory(modl) is not None]
    if not is_empty(userfactory):
        return userfactory[0]
    return None

def format_number(number, is_pourcent=False, decimal=2):
    number = number if number else 0
    return f'{number:,.{decimal}{"%" if is_pourcent else "f"}}'

# https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable?page=1&tab=votes#tab-top
def datetime_converter(o):
    if isinstance(o, (datetime.date, datetime.datetime, datetime.time)):
        return o.isoformat()


from dataclasses import dataclass
from typing import Optional

from .enums import ButtonLevel, ButtonType


@dataclass(unsafe_hash=True)
class ButtonConfig:
    label: Optional[str] = None
    icon: Optional[str] = None
    title: Optional[str] = None

    level: ButtonLevel = ButtonLevel.DEFAULT

    weight: int = 100

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        assert self.label or self.icon, "Either label or icon has to be defined."

    def __iter__(self):
        if hasattr(super(), "__iter__"):
            yield from super().__iter__()

        for key in ["label", "icon", "title"]:
            value = getattr(self, key, None)
            if value:
                yield key, value
        yield "level", self.level.value


@dataclass(unsafe_hash=True)
class ButtonTypeMixin:
    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        assert hasattr(self, "button_type"), "button_type cannot be None."

    def __iter__(self):
        if hasattr(super(), "__iter__"):
            yield from super().__iter__()

        yield "type", self.button_type.value


@dataclass(unsafe_hash=True)
class ButtonUrlMixin:
    key: Optional[str] = None
    endpoint: Optional[str] = None

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        assert bool(self.key) != bool(self.endpoint), "Either key or endpoint has to be defined. (Not both)"

    def __iter__(self):
        if hasattr(super(), "__iter__"):
            yield from super().__iter__()

        if self.key:
            yield "key", self.key

        if self.endpoint:
            yield "endpoint", self.endpoint





def historical_models(historical_app_models):
    models = list()
    for app in historical_app_models:
        models.extend(list(map(lambda model: f"{app[0]}.{model}", app[1])))
    return models



# Generated by Django 3.0.6 on 2020-05-14 15:05
import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0006_auto_20200507_0913"),
    ]

    operations = [
        migrations.CreateModel(
            name="Asset",
            fields=[
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False,),),
                ("file", models.FileField(upload_to="markdown/assets")),
            ],
            options={"verbose_name": "Asset", "verbose_name_plural": "Assets",},
        ),
    ]



# import pytest
# import time
# from selenium import webdriver
# from channels.testing import ChannelsLiveServerTestCase

# from django.contrib.auth import get_user_model, authenticate
# from .selenium.utils import menu_items_for_user

# def login(url, username, password):
#     driver = webdriver.Chrome()
#     driver.get(f"{url}/frontend/")
#     driver.implicitly_wait(10)
#     driver.find_element_by_xpath("//input[contains(@class, 'form-control')]").send_keys(username)
#     driver.find_element_by_xpath("//input[contains(@class, 'text-input')]").send_keys(password)
#     driver.find_element_by_xpath("//button").click()
#     time.sleep(1)
#     return driver


# @pytest.mark.django_db
# class TestLive(ChannelsLiveServerTestCase):

#     @pytest.mark.parametrize
#     def test_something1(self):
#         user = get_user_model().objects.create_superuser(
#             username="root",
#             email="a@a.de",
#             password="root"
#         )
#         driver = login(self.live_server_url, "root", "root")

#         # self.driver.find_element_by_class_name("root-menu__trigger").click()

#         # menu_items = self.driver.find_elements_by_class_name("menu__menu-item")
#         # for menu_item in menu_items:
#         #     menu_item.click()
#         #     time.sleep(3)

#         assert False


from typing import Optional

from rest_framework.request import Request
from bridger.metadata.mixins import BridgerViewSetConfig


class PreviewConfig(BridgerViewSetConfig):
    DISPLAY_TYPE = "instance_display"

    def get_buttons(self):
        return []

    def _get_buttons(self):
        return [dict(button) for button in self.get_buttons()]

    def get_display(self):
        return None

    def _get_display(self):
        if display := self.get_display():
            if self.DISPLAY_TYPE == "instance_display":
                return list(display)
            return display
        return None

    def _get_display_type(self):
        return self.DISPLAY_TYPE

    def get_metadata(self):
        buttons = self._get_buttons()
        display = self._get_display()

        if len(buttons) > 0 or display is not None:
            yield "buttons", buttons
            yield "display", display
            yield "type", self._get_display_type()

    @classmethod
    def get_metadata_key(cls):
        return "preview"

    @classmethod
    def get_metadata_fieldname(cls):
        return "preview_config_class"


class PreviewConfigMixin:
    preview_config_class = PreviewConfig



from bridger import filters

from .models import Tag


class TagFilterMixin(filters.FilterSet):

    tags = filters.ModelMultipleChoiceFilter(
        endpoint="bridger:tagrepresentation-list", value_key="id", label_key="title", queryset=Tag.objects.all(),
    )



# Generated by Django 2.2.10 on 2020-02-18 12:22

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("bridger", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID",),),
                ("title", models.CharField(max_length=512)),
                ("message", models.TextField(blank=True, null=True)),
                ("buttons", models.JSONField(blank=True, default=dict, null=True),),
                ("timestamp_created", models.DateTimeField(auto_now_add=True)),
                ("timestamp_received", models.DateTimeField(blank=True, null=True)),
                ("timestamp_read", models.DateTimeField(blank=True, null=True)),
                ("timestamp_mailed", models.DateTimeField(blank=True, null=True)),
                (
                    "send_type",
                    models.CharField(
                        choices=[("system", "System"), ("mail", "Mail"), ("system_and_mail", "System and Mail"),],
                        default="system",
                        max_length=32,
                    ),
                ),
                (
                    "recipient",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="notifications", to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from datetime import timedelta
from typing import Dict, List

from django.conf import settings
from django.contrib.auth import get_user_model
from jwt import decode as jwt_decode
from rest_framework import authentication
from rest_framework.request import Request
from rest_framework.reverse import reverse
from rest_framework_simplejwt.exceptions import AuthenticationFailed, InvalidToken, TokenError
from rest_framework_simplejwt.tokens import UntypedToken


def unauthenticated(request: Request) -> Dict:
    return {"type": None}


def get_dev_user_from_settings(username_field_key: str) -> List[Dict[str, str]]:
    if dev_user := settings.DEV_USER:
        users = list()
        for user in dev_user.split(","):
            username, password = user.split(":")
            users.append({username_field_key: username, "password": password})

        return users
    return []


def jwt_auth(request: Request) -> Dict:
    user_model = get_user_model()
    username_field_key = user_model.USERNAME_FIELD
    username_field_label = user_model._meta.get_field(username_field_key).verbose_name

    access = settings.SIMPLE_JWT.get("ACCESS_TOKEN_LIFETIME", timedelta(minutes=5))
    refresh = settings.SIMPLE_JWT.get("REFRESH_TOKEN_LIFETIME", timedelta(days=1))

    config = {
        "type": "JWT",
        "config": {
            "token": reverse("token_obtain_pair", request=request),
            "refresh": reverse("token_refresh", request=request),
            "verify": reverse("token_verify", request=request),
            "token_lifetime": {"access": access, "refresh": refresh},
            "username_field_key": username_field_key,
            "username_field_label": username_field_label,
        },
    }

    if users := get_dev_user_from_settings(username_field_key):
        config["dev_users"] = users
    return config


class JWTCookieAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request: Request):
        try:
            jwt_access_token = request.COOKIES["JWT-access"]
            UntypedToken(jwt_access_token)
            decoded_data = jwt_decode(jwt_access_token, settings.SECRET_KEY, algorithms=["HS256"])
            user = get_user_model().objects.get(id=decoded_data["user_id"])
            return (user, None)
        except (InvalidToken, TokenError, KeyError, get_user_model().DoesNotExist):
            raise AuthenticationFailed("Authentication Failed.")
            return None



from bridger.buttons.metadata_config import ButtonConfigMixin
from bridger.display.metadata_config import DisplayConfigMixin
from bridger.titles.metadata_config import TitleConfigMixin
from bridger.endpoints.metadata_config import EndpointConfigMixin
from bridger.preview.metadata_config import PreviewConfigMixin

from .metadata import BridgerMetadata


class MetadataMixin(DisplayConfigMixin, EndpointConfigMixin, TitleConfigMixin, ButtonConfigMixin, PreviewConfigMixin):
    metadata_class = BridgerMetadata



from django.contrib import admin

from .markdown.admin import AssetModelAdmin
from .models import FrontendUserConfiguration
from .notifications.admin import NotificationModelAdmin
from .tags.admin import TagModelAdmin
import csv
from django import forms
import pandas as pd
from django.http import HttpResponse
from django.urls import path
from django.shortcuts import render, redirect
from io import StringIO
import numpy as np

class FrontendUserConfigurationInline(admin.TabularInline):
    model = FrontendUserConfiguration
    extra = 0
    fields = ["id", "user", "config"]
    readonly_fields = ["id", "user", "config"]


@admin.register(FrontendUserConfiguration)
class FrontendUserConfigurationModelAdmin(admin.ModelAdmin):
    list_display = ["id", "user"]
    fieldsets = (("Main Information", {"fields": ("user", "parent_configuration", "config")}),)
    inlines = [FrontendUserConfigurationInline]


class CsvImportForm(forms.Form):
    csv_file = forms.FileField()

class ExportCsvMixin:
    def export_as_csv(self, admin, request, queryset, *args, **kwargs):
        meta = queryset.model._meta
        field_names = [field.name for field in meta.fields]

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={}.csv".format(meta)
        writer = csv.writer(response)

        writer.writerow(field_names)
        for obj in queryset:
            row = writer.writerow([getattr(obj, field) for field in field_names])

        return response

    def get_actions(self, request):
        actions = super().get_actions(request)
        actions["export_as_csv"] = (self.export_as_csv, "export_as_csv", "Export Selected")
        return actions

class ImportCsvMixin:
    change_list_template = 'bridger/admin/change_list.html'

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [path("import-csv/", self._import_csv)]
        return my_urls + urls

    def manipulate_df(self, df):
        return df
    
    def process_model(self, model):
        self.model.create(**model)
    
    def get_import_fields(self):
        return [f.name for f in self.model._meta.get_fields()]

    def _import_csv(self, request):
        if request.method == "POST":
            csv_file = request.FILES["csv_file"]

            str_text = ""
            for line in csv_file:
                str_text = str_text + line.decode()
            # Import csv as df
            df = pd.read_csv(StringIO(str_text))
            # Sanitize dataframe
            df = df.where(pd.notnull(df), None)
            df = df.drop(df.columns.difference(self.get_import_fields()), axis=1)

            # Overide this function if there is foreign key ids in the dataframe
            df = self.manipulate_df(df)
            errors = 0
            for model in df.to_dict("records"):
                #by default, process the modela as a create request. Can be override to change the behavior
                try:
                    self.process_model(model)
                except Exception as e:
                    errors += 1
                    pass
            self.message_user(request, f"Your csv file has been imported ( {df.shape[0] -errors }  imported, {errors} errors)")
            return redirect("..")
        form = CsvImportForm()
        payload = {"form": form}
        return render(request, "bridger/admin/csv_form.html", payload)




def enumerated_string_join(l):
    ''' Generates a human readable string out of a list '''

    if len(l) > 0:
        l = [str(i) for i in l]
        return ' and '.join([', '.join(l[:-1]), l[-1]]) if len(l) > 1 else l[0]
    return ''

def format_number(number, is_percent=False, decimal=2):
    number = number if number else 0
    return f'{number:,.{decimal}{"%" if is_percent else "f"}}'
    
class ReferenceIDMixin:
    @property
    def reference_id(self):
        return f"{self.id:06}"


from django_filters.rest_framework import DjangoFilterBackend as StandardFilterBackend

from .filterset import FilterSet


class DjangoFilterBackend(StandardFilterBackend):
    filterset_base = FilterSet

    def get_filterset_class(self, view, queryset=None):

        if method := getattr(view, "get_filterset_class", None):
            view.filterset_class = method(view.request)

        if method := getattr(view, "get_filterset_fields", None):
            view.filterset_fields = method(view.request)

        return super().get_filterset_class(view, queryset)



from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from bridger.settings import bridger_settings


class ShareAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request: Request) -> Response:
        serializer = bridger_settings.DEFAULT_SHARE_SERIALIZER(data=request.data)
        serializer.is_valid(raise_exception=True)
        bridger_settings.DEFAULT_SHARE_NOTIFICATION(**request.data, user=request.user)
        return Response({})



from contextlib import suppress
from typing import List, Set, Dict

from rest_framework.request import Request

from bridger.enums import Button
from bridger.metadata.mixins import BridgerViewSetConfig
from bridger.signals.instance_buttons import add_instance_button
from bridger.buttons.buttons import DropDownButton

class ButtonConfig(BridgerViewSetConfig):

    # Utils
    def order_buttons(self, buttons: Set, ordering: List) -> List:
        yield from filter(lambda b: b in buttons, ordering)

    # FSM Button Configuration
    FSM_LIST = True
    FSM_INSTANCE = True
    FSM_DROPDOWN = False
    FSM_DROPDOWN_ICON = "wb-icon-plus"
    FSM_DROPDOWN_LABEL = "Transitions"
    FSM_WEIGHT = 100

    def get_fsm_buttons(self) -> Set:
        if self.FSM_DROPDOWN and len(self.view.FSM_BUTTONS) > 0:
            return {DropDownButton(
                label=self.FSM_DROPDOWN_LABEL,
                icon=self.FSM_DROPDOWN_ICON,
                title=self.FSM_DROPDOWN_LABEL,
                weight=self.FSM_WEIGHT,
                buttons=tuple(self.view.FSM_BUTTONS),
            )}
        return getattr(self.view, "FSM_BUTTONS", set())


    # List Button Configuration
    LIST_BUTTONS = {Button.NEW.value, Button.REFRESH.value}
    LIST_BUTTONS_ORDERING = [Button.NEW.value, Button.REFRESH.value]

    def get_list_buttons(self) -> List:
        content_type = self.view.get_content_type()
        add_permission = f"{content_type.app_label}.add_{content_type.model}"

        if not self.request.user.has_perm(add_permission):
            with suppress(KeyError):
                self.LIST_BUTTONS.remove(Button.NEW.value)

        yield from self.order_buttons(self.LIST_BUTTONS, self.LIST_BUTTONS_ORDERING)

    # Instance Button Configuration
    INSTANCE_BUTTONS = {Button.SAVE.value, Button.REFRESH.value, Button.DELETE.value}
    INSTANCE_BUTTONS_ORDERING = [Button.SAVE.value, Button.REFRESH.value, Button.DELETE.value]

    def get_instance_buttons(self) -> List:
        content_type = self.view.get_content_type()
        change_permission = f"{content_type.app_label}.change_{content_type.model}"
        delete_permission = f"{content_type.app_label}.delete_{content_type.model}"

        if not self.request.user.has_perm(change_permission):
            with suppress(KeyError):
                self.INSTANCE_BUTTONS.remove(Button.SAVE.value)

        if not self.request.user.has_perm(delete_permission):
            with suppress(KeyError):
                self.INSTANCE_BUTTONS.remove(Button.DELETE.value)

        yield from self.order_buttons(self.INSTANCE_BUTTONS, self.INSTANCE_BUTTONS_ORDERING)

    # Create Button Configuration
    CREATE_BUTTONS = {Button.SAVE.value, Button.SAVE_AND_CLOSE.value, Button.SAVE_AND_NEW.value, Button.RESET.value}
    CREATE_BUTTONS_ORDERING = [Button.SAVE.value, Button.SAVE_AND_CLOSE.value, Button.SAVE_AND_NEW.value, Button.RESET.value]

    def get_create_buttons(self) -> List:
        yield from self.order_buttons(self.CREATE_BUTTONS, self.CREATE_BUTTONS_ORDERING)

    # Custom Instance Button Configuration
    CUSTOM_INSTANCE_BUTTONS = set()
    CUSTOM_LIST_INSTANCE_BUTTONS = set()

    def get_custom_instance_buttons(self) -> Set:
        return self.CUSTOM_INSTANCE_BUTTONS

    def get_custom_list_instance_buttons(self) -> Set:
        return self.CUSTOM_LIST_INSTANCE_BUTTONS

    def apply_request(self, buttons):
        for button in buttons:
            button.request = self.request
            if hasattr(button, "buttons"):
                self.apply_request(button.buttons)

    def _get_custom_instance_buttons(self):
        custom_instance_buttons = set()
        if self.instance:
            custom_instance_buttons |= self.get_custom_instance_buttons()
            if self.FSM_INSTANCE:
                custom_instance_buttons |= self.get_fsm_buttons()

        else:
            custom_instance_buttons |= self.get_custom_list_instance_buttons()
            if self.FSM_LIST:
                custom_instance_buttons |= self.get_fsm_buttons()

        remote_buttons = add_instance_button.send(self.view.__class__, many=True)
        custom_instance_buttons |= set([button for _, button in remote_buttons])

        self.apply_request(custom_instance_buttons)

        iter_key = lambda e: (e.weight, e.label, e.title)
        for element in sorted(custom_instance_buttons, key=iter_key):
            yield dict(element)

    # Custom Button Configuration
    CUSTOM_BUTTONS = set()

    def get_custom_buttons(self) -> Set:
        return self.CUSTOM_BUTTONS

    def _get_custom_buttons(self):
        custom_buttons = self.get_custom_buttons()
        self.apply_request(custom_buttons)

        iter_key = lambda e: (e.weight, e.label, e.title)
        for element in sorted(custom_buttons, key=iter_key):
            yield dict(element)

    def get_metadata(self) -> Dict:
        yield "custom_instance", self._get_custom_instance_buttons()
        yield "custom", self._get_custom_buttons()

    @classmethod
    def get_metadata_key(cls):
        return "buttons"

    @classmethod
    def get_metadata_fieldname(cls):
        return "button_config_class"

class ButtonConfigMixin:
    button_config_class = ButtonConfig



# Generated by Django 3.0.6 on 2020-06-09 11:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0011_asset_file_url_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Tag",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID",),),
                ("title", models.CharField(max_length=255, unique=True)),
                ("color", models.CharField(default="#D3D3D3", max_length=7)),
            ],
            options={"verbose_name": "Tag", "verbose_name_plural": "Tags",},
        ),
    ]



# Generated by Django 2.2.12 on 2020-04-02 15:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0004_auto_20200327_1152"),
    ]

    operations = [
        migrations.AlterField(model_name="notification", name="endpoint", field=models.URLField(blank=True, null=True),),
    ]



from enum import Enum


class ButtonLevel(Enum):
    SUCCESS = "success"
    ERROR = "error"
    WARNING = "warning"
    LINK = "link"
    DEFAULT = "default"


class ButtonType(Enum):
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"


class HyperlinkTarget(Enum):
    BLANK = "_blank"
    SELF = "_self"
    PARENT = "_parent"
    TOP = "_top"



# Generated by Django 3.0.14 on 2021-05-10 08:48

from slugify import slugify
from django.db import migrations, models


def populate_slug_field(apps, schema_editor):
    Tag = apps.get_model("bridger", "Tag")
    for tag in Tag.objects.all():
        tag.slug = slugify(tag.title)
        tag.save()


class Migration(migrations.Migration):

    dependencies = [
        ('bridger', '0012_tag'),
    ]

    operations = [
        migrations.AddField(
            model_name='tag',
            name='slug',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.RunPython(populate_slug_field),
    ]



# Generated by Django 3.0.6 on 2020-05-14 15:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0008_auto_20200514_1711"),
    ]

    operations = [
        migrations.RenameModel(old_name="Asset111", new_name="Asset",),
    ]



from enum import Enum

import celery
from django.contrib.auth import get_user_model

from django.db import models, transaction
from django.db.models.signals import post_save
from django.dispatch import receiver
from bridger.buttons import WidgetButton
from django.conf import settings

class NotificationSendType(Enum):
    SYSTEM = "system"
    MAIL = "mail"
    SYSTEM_AND_MAIL = "system_and_mail"


class Notification(models.Model):
    """
        A model that represents a notification that is send as a system notification or mail or both
    """

    recipient = models.ForeignKey(to=get_user_model(), related_name="notifications", on_delete=models.CASCADE)

    title = models.CharField(max_length=512)
    message = models.TextField(null=True, blank=True)
    buttons = models.JSONField(default=list, null=True, blank=True)

    endpoint = models.URLField(null=True, blank=True, max_length=2048)

    timestamp_created = models.DateTimeField(auto_now_add=True, verbose_name="Created")
    timestamp_received = models.DateTimeField(null=True, blank=True, verbose_name="Received")
    timestamp_read = models.DateTimeField(null=True, blank=True, verbose_name="Read")
    timestamp_mailed = models.DateTimeField(null=True, blank=True, verbose_name="Mailed")

    send_type_choices = (
        (NotificationSendType.SYSTEM.value, "System"),
        (NotificationSendType.MAIL.value, "Mail"),
        (NotificationSendType.SYSTEM_AND_MAIL.value, "System and Mail"),
    )

    send_type = models.CharField(max_length=32, choices=send_type_choices, default=NotificationSendType.SYSTEM.value,)

    def __str__(self):
        return f"{self.recipient} {self.title}"

    def to_payload(self):
        return {
            "id": self.id,
            "title": self.title,
            "message": self.message,
            "buttons": self.buttons,
            "endpoint": self.endpoint,
        }
    def get_endpoint_basename(cls):
        return "bridger:notification"

@receiver(post_save, sender=Notification)
def post_create_notification(sender, instance, created, **kwargs):
    from .tasks import send_mail, send_system
    if created:
        dispatch = {
            NotificationSendType.MAIL.value: [send_mail],
            NotificationSendType.SYSTEM.value: [send_system],
            NotificationSendType.SYSTEM_AND_MAIL.value: [send_system, send_mail],
        }
        for action in dispatch[instance.send_type]:
            try:
                action.apply_async((instance.id,), countdown=getattr(settings, "BRIDGER_NOTIFICATION_DELAY_SECONDS", 10))
            except Exception as e:
                pass
            # transaction.on_commit(lambda: celery.execute.send_task(action, args=[instance.id]))



from rest_framework.generics import GenericAPIView as OriginalGenericAPIView


class GenericAPIView(OriginalGenericAPIView):
    def get_serializer_changes(self, serializer):
        return serializer

    def get_serializer(self, *args, **kwargs):
        return self.get_serializer_changes(super().get_serializer(*args, **kwargs))
